//- -----------------------------------------------------------------------------------------------------------------------
//- HomeMatic - Gardena - Ventil
//- HM_LC_SW1_BA_PCB_GAR1251
//- Version: 3.3
//- Date: 2020-05-29
//- -----------------------------------------------------------------------------------------------------------------------
// AskSin++
// 2016-10-31 papa            Creative Commons - http://creativecommons.org/licenses/by-nc-sa/3.0/de/
// HB-UNI-Sensor1
// 2019-02-01 Tom Major       Creative Commons - https://creativecommons.org/licenses/by-nc-sa/4.0/
//
// 2020-02-21 Wolfram Winter  Creative Commons - https://creativecommons.org/licenses/by-nc-sa/4.0/
//                            CCU-Einstellungen (Batterie-Schwelle, Geräte-LED) + BatterieCheck + CC1101 Frequenz
// 2020-05-29 Wolfram Winter  Creative Commons - https://creativecommons.org/licenses/by-nc-sa/4.0/
//                            Konfiguration für Platine mit 'echter' Batteriespannungsmessung
//
// You are free to Share & Adapt under the following terms:
// Give Credit, NonCommercial, ShareAlike
//- -----------------------------------------------------------------------------------------------------------------------

// define this to read the device id, serial and device type from bootloader section
// #define USE_OTA_BOOTLOADER

#define USE_WOR
#define EI_NOTEXTERNAL
#include <EnableInterrupt.h>
#include <AskSinPP.h>
#include <LowPower.h>
#include <MultiChannelDevice.h>
#include <Switch.h>

#include "tmBattery.h"
#include "HM_LC_SW1_BA_PCB_GAR1251.h"

#ifdef CLOCK_SYSCLOCK
#define CLOCK sysclock
#define SAVEPWR_MODE Sleep<>
#elif defined CLOCK_RTC
#define CLOCK rtc
#define SAVEPWR_MODE SleepRTC
#undef seconds2ticks
#define seconds2ticks(tm) (tm)
#else
#error INVALID CLOCK OPTION
#endif

// number of available peers per channel
#define PEERS_PER_CHANNEL 8

// all library classes are placed in the namespace 'as'
using namespace as;

// Define all device properties
const struct DeviceInfo PROGMEM devinfo = {
    cDEVICE_ID,              // Device ID
    cDEVICE_SERIAL,          // Device Serial
    cDEVICE_MODEL ,          // Device Model
    cDEVICE_FIRMWARE,        // Firmware Version
    as::DeviceType::Switch,  // Device Type
    { 0x01, 0x00 }           // Info Bytes
};

/**
 * Configure the used hardware
 */
typedef AvrSPI<CC1101_CS_PIN, CC1101_MOSI_PIN, CC1101_MISO_PIN, CC1101_SCK_PIN>  SPIType;
typedef Radio<SPIType, CC1101_GDO0_PIN>                                          RadioType;
typedef StatusLed<LED_PIN>                                                       LedType;
typedef AskSin<LedType, BAT_SENSOR, RadioType>                                   BaseHal;

class Hal : public BaseHal {
public:
    void init(const HMID& id)
    {
        BaseHal::init(id);
#ifdef CLOCK_RTC
        rtc.init();    // init real time clock - 1 tick per second
#endif
        // measure battery every x h
        battery.init(seconds2ticks(BAT_CHECK_TIME), CLOCK);
        battery.low(BAT_VOLT_LOW);
        battery.critical(BAT_VOLT_CRITICAL);
    }

    bool runready() { return CLOCK.runready() || BaseHal::runready(); }
} hal;

DEFREGISTER(Reg0,DREG_INTKEY,DREG_LEDMODE,MASTERID_REGS,DREG_LOWBATLIMIT)
class SwList0 : public RegList0<Reg0> {
public:
    SwList0(uint16_t addr) : RegList0<Reg0>(addr) {}
    void defaults () {
      clear();
      lowBatLimit(BAT_VOLT_LOW);
      ledMode(cDEVICE_LED_MODE);
    }
};

// this class controls the pin high/low sequence
class SwAlarm : public Alarm {
private:
    uint8_t pin;
    uint8_t time;
    uint8_t state;
public:
    SwAlarm () : Alarm(0), pin(0), time(0), state(0) {
      async(true);
    }
    virtual ~SwAlarm () {}

    virtual void trigger (AlarmClock& clock) {
      switch(state) {
      case 1:
        set(millis2ticks(500));
        digitalWrite(ENABLE_PIN,HIGH);
        break;
      case 2:
        set(millis2ticks(time));
        digitalWrite(pin,HIGH);
        break;
      case 3:
        set(millis2ticks(100));
        digitalWrite(pin,LOW);
        break;
      case 4:
        digitalWrite(ENABLE_PIN,LOW);
        state=0;
        break;
      }
      if( state != 0 ) {
        ++state;
        clock.add(*this);
      }
    }

    bool press (uint8_t p,uint8_t t,AlarmClock& clock) {
      // check that we are not running
      if( state == 0 ) {
        pin = p;
        time = t;
        state = 1;
        trigger(clock);
        return true;
      }
      return false;
    }
};

class SwChannel : public SwitchChannel<Hal, PEERS_PER_CHANNEL, SwList0> {
protected:
    uint8_t pinoff;
    SwAlarm swalarm;
public:
    typedef SwitchChannel<Hal, PEERS_PER_CHANNEL, SwList0> BaseChannel;
    virtual ~SwChannel () {}

    void init (uint8_t pon,uint8_t poff) {
      BaseChannel::init(pon,false);
      pinoff = poff;
      pinMode(pinoff,OUTPUT);
 
      DPRINT("INIT Pin "); DDEC(pinoff); DPRINTLN(" - OUTPUT");

    }

    // replace standard switch method - trigger alarm to start sequence
    // virtual void switchState(__attribute__((unused)) uint8_t oldstate,uint8_t newstate) {
    virtual void switchState(__attribute__((unused)) uint8_t oldstate,uint8_t newstate,__attribute__((unused)) uint32_t delay) {
      bool pressed = false;

      DPRINTLN("### SwitchChannel - switchState");
      
      if( newstate == AS_CM_JT_ON ) {
        DPRINTLN("SwitchChannel - AS_CM_JT_ON");
        pressed = swalarm.press(pin,250,sysclock);
      }
      else if ( newstate == AS_CM_JT_OFF ) {
        DPRINTLN("SwitchChannel - AS_CM_JT_OFF");
        pressed = swalarm.press(pinoff,62,sysclock);
      }
      if( pressed == true ) {
        DPRINTLN("SwitchChannel - pressed");
        BaseChannel::changed(true);
      }
    }
};

// setup the device with channel type and number of channels
class SwitchType : public MultiChannelDevice<Hal,SwChannel,2,SwList0> {
public:
    typedef MultiChannelDevice<Hal,SwChannel,2,SwList0> DevType;
    SwitchType (const DeviceInfo& i,uint16_t addr) : DevType(i,addr) {}
    virtual ~SwitchType () {}

    bool init (Hal& hal) {
      bool rtn = DevType::init(hal);
#ifdef cRESET_ONSTART
      DPRINTLN("Device value reset");
      // Wert der Batterie-Warnung beim Start immer auf den Anfangswert zurücksetzen 
      this->getList0().lowBatLimit(BAT_VOLT_LOW);
      DPRINT("*Set LOW BAT Limit: ");
      DDECLN(this->getList0().lowBatLimit());
      // Wert des LED-Modus beim Start immer auf den Anfangswert zurücksetzen 
      this->getList0().ledMode(cDEVICE_LED_MODE);
      DPRINT("*Set LED-Mode: "); 
      DDECLN(this->getList0().ledMode());
#endif
      return(rtn);
    }

    virtual void configChanged () {
      DevType::configChanged();
      uint8_t lowBat = this->getList0().lowBatLimit();
#if BAT_SENSOR_MODE == 1
      lowBat = BAT_VOLT_LOW;
#else
      if( (lowBat < BAT_VOLT_CRITICAL) || (lowBat > BAT_VOLT_LOW_MAX) || (lowBat <= 0) ) {
        lowBat = BAT_VOLT_LOW;
      }
#endif
      DPRINT("*Set LOW BAT Limit: ");
      DDEC(lowBat);
      DPRINT(" (got: ");
      DDEC(this->getList0().lowBatLimit());
      DPRINTLN(")");
      this->battery().low(lowBat);
      DPRINT("*Set LED-Mode: "); 
      DDECLN(this->getList0().ledMode());
      this->led().set( (LedStates::Mode) (this->getList0().ledMode()) );
    }
};

SwitchType sdev(devinfo,0x20);
#if cDEVICE_CB_MODE == 0
ConfigButton<SwitchType> cfgBtn(sdev);
#else
ConfigToggleButton<SwitchType> cfgBtn(sdev);
#endif

#ifndef USE_WOR
    BurstDetector<Hal> bd(hal);
#endif

void initPeerings (bool first) {
    // create internal peerings - CCU2 needs this
    if( first == true ) {
      HMID devid;
      sdev.getDeviceID(devid);
      for( uint8_t i=1; i<=sdev.channels(); ++i ) {
        Peer ipeer(devid,i);
        sdev.channel(i).peer(ipeer);
      }
    }
}

void setup () {
    DINIT(57600,ASKSIN_PLUS_PLUS_IDENTIFIER);
    bool first = sdev.init(hal);

    pinMode(ENABLE_PIN,OUTPUT);
    digitalWrite(ENABLE_PIN,LOW);

    sdev.channel(1).init(RELAY1_ON_PIN, RELAY1_OFF_PIN);
    sdev.channel(2).init(RELAY2_ON_PIN, RELAY2_OFF_PIN);

    buttonISR(cfgBtn,CONFIG_BUTTON_PIN);

    initPeerings(first);

#ifndef USE_WOR
    // start burst detection
    bd.enable(sysclock);
#endif

#ifdef USE_CC1101_INIT
    // Set own frequency for CC1101
    // If other values are found in EEPROM, these are overwritten
    hal.radio.initReg(CC1101_FREQ2, CC1101_FREQ2_VAL);
    hal.radio.initReg(CC1101_FREQ1, CC1101_FREQ1_VAL);
    hal.radio.initReg(CC1101_FREQ0, CC1101_FREQ0_VAL);
    DPRINT("*Set Config Freq: 0x");
    DHEX((uint8_t)(CC1101_FREQ2_VAL));
    DHEX((uint8_t)(CC1101_FREQ1_VAL));
    DHEXLN((uint8_t)(CC1101_FREQ0_VAL));
#endif

    // stay on for 15 seconds after start
    hal.activity.stayAwake(seconds2ticks(15));

    // measure battery every BAT_CHECK_TIME
    // hal.battery.init(seconds2ticks(BAT_CHECK_TIME),sysclock);
    // class Hal by default: measure battery every BAT_CHECK_TIME

    sdev.initDone();

    DPRINTLN(F("Setup done"));
    // Output ID and Serial in serial console
    DDEVINFO(sdev);
#ifdef CLOCK_SYSCLOCK
    DPRINTLN(F("Clock SYSCLOCK"));
#elif defined CLOCK_RTC
    DPRINTLN(F("Clock RTC"));
#endif

}

void loop() {
    bool worked = hal.runready();
    bool poll = sdev.pollRadio();
    if ( worked == false && poll == false ) {
      // deep discharge protection
      // if we drop below critical battery level - switch off all and sleep forever
      if (hal.battery.critical()) {
        // this call will never return
        hal.activity.sleepForever(hal);
      }
      // if nothing to do - go sleep
      hal.activity.savePower<Sleep<> >(hal);
    }
}
